"""Enhanced workflow execution with logging and validation."""

import json
import time
from datetime import datetime
from typing import Dict, Any, Optional, Generator, Union, Callable, List, AsyncGenerator
from enum import Enum
import asyncio
import traceback
import uuid

from .client import execute_workflow as _execute_workflow_raw
from .core.exceptions import WorkflowValidationError


class LogLevel(str, Enum):
    """Logging levels for workflow execution."""
    MINIMAL = "minimal"    # Only major events
    NORMAL = "normal"      # Standard logging
    VERBOSE = "verbose"    # All events with details
    DEBUG = "debug"        # Full debugging output


class ExecutionMode(str, Enum):
    """Execution modes for workflow execution."""
    RAW = "raw"           # Raw SSE events
    LOGGING = "logging"   # Enhanced logging output
    EVENTS = "events"     # Structured event objects


def validate_workflow_definition(workflow: Dict[str, Any]) -> List[str]:
    """Validate workflow definition before execution.
    
    Args:
        workflow: Workflow definition to validate
        
    Returns:
        List of validation errors (empty if valid)
    """
    errors = []
    
    # Basic structure validation
    if not isinstance(workflow, dict):
        errors.append("Workflow definition must be a dictionary")
        return errors
    
    # Required fields
    if "name" not in workflow:
        errors.append("Workflow must have a 'name' field")
    elif not workflow["name"] or not workflow["name"].strip():
        errors.append("Workflow name cannot be empty")
    
    if "steps" not in workflow:
        errors.append("Workflow must have a 'steps' field")
    elif not isinstance(workflow["steps"], list):
        errors.append("Workflow 'steps' must be a list")
    elif len(workflow["steps"]) == 0:
        errors.append("Workflow must have at least one step")
    
    # Validate steps
    step_names = set()
    for i, step in enumerate(workflow.get("steps", [])):
        if not isinstance(step, dict):
            errors.append(f"Step {i} must be a dictionary")
            continue
        
        # Step name validation
        if "name" not in step:
            errors.append(f"Step {i} must have a 'name' field")
        else:
            step_name = step["name"]
            if not step_name or not step_name.strip():
                errors.append(f"Step {i} name cannot be empty")
            elif step_name in step_names:
                errors.append(f"Duplicate step name '{step_name}' found")
            step_names.add(step_name)
        
        # Executor validation
        if "executor" in step:
            executor = step["executor"]
            if not isinstance(executor, dict):
                errors.append(f"Step {i} executor must be a dictionary")
            elif "type" not in executor:
                errors.append(f"Step {i} executor must have a 'type' field")
            elif executor["type"] == "tool" and "config" in executor:
                config = executor["config"]
                if "tool_def" in config:
                    tool_def = config["tool_def"]
                    
                    # Validate bounded services if present
                    if "with_services" in tool_def:
                        services = tool_def["with_services"]
                        if not isinstance(services, list):
                            errors.append(f"Step {i} tool_def with_services must be a list")
                        else:
                            service_names = set()
                            for j, service in enumerate(services):
                                if not isinstance(service, dict):
                                    errors.append(f"Step {i} service {j} must be a dictionary")
                                    continue
                                
                                # Required service fields
                                if "name" not in service:
                                    errors.append(f"Step {i} service {j} must have a 'name' field")
                                elif service["name"] in service_names:
                                    errors.append(f"Step {i} has duplicate service name '{service['name']}'")
                                else:
                                    service_names.add(service["name"])
                                
                                if "image" not in service:
                                    errors.append(f"Step {i} service {j} must have an 'image' field")
                                
                                # Validate ports
                                if "exposed_ports" in service:
                                    ports = service["exposed_ports"]
                                    if not isinstance(ports, list):
                                        errors.append(f"Step {i} service {j} exposed_ports must be a list")
                                    else:
                                        for port in ports:
                                            if not isinstance(port, int) or port < 1 or port > 65535:
                                                errors.append(f"Step {i} service {j} invalid port: {port}")
        
        # Dependencies validation
        if "depends" in step:
            depends = step["depends"]
            if isinstance(depends, str):
                if depends not in step_names and depends != step.get("name"):
                    errors.append(f"Step {i} depends on unknown step '{depends}'")
            elif isinstance(depends, list):
                for dep in depends:
                    if dep not in step_names and dep != step.get("name"):
                        errors.append(f"Step {i} depends on unknown step '{dep}'")
    
    return errors


def execute_workflow_with_logging(
    workflow_definition: Union[Dict[str, Any], str],
    api_key: str,
    mode: ExecutionMode = ExecutionMode.LOGGING,
    log_level: LogLevel = LogLevel.NORMAL,
    validate: bool = True,
    on_event: Optional[Callable[[Dict[str, Any]], None]] = None,
    parameters: Optional[Dict[str, Any]] = None,
    base_url: str = "https://api.kubiya.ai"
) -> Union[Generator[str, None, None], Generator[Dict[str, Any], None, None]]:
    """Execute workflow with enhanced logging and validation.
    
    Args:
        workflow_definition: Workflow definition (dict or JSON string)
        api_key: Kubiya API key
        mode: Execution mode (raw, logging, events)
        log_level: Logging verbosity level
        validate: Whether to validate workflow before execution
        on_event: Optional callback for each event
        parameters: Workflow parameters
        base_url: Base URL for the Kubiya API
        
    Returns:
        Generator yielding events based on mode
        
    Raises:
        WorkflowValidationError: If validation fails
        ValueError: For invalid arguments
    """
    # Convert string to dict if needed
    if isinstance(workflow_definition, str):
        try:
            workflow_definition = json.loads(workflow_definition)
        except json.JSONDecodeError as e:
            raise WorkflowValidationError(f"Invalid workflow JSON: {str(e)}")
    
    # Validate workflow if requested
    if validate:
        errors = validate_workflow_definition(workflow_definition)
        if errors:
            raise WorkflowValidationError(f"Workflow validation failed: {'; '.join(errors)}")
    
    # Raw mode - just pass through
    if mode == ExecutionMode.RAW:
        yield from _execute_workflow_raw(
            workflow_definition=workflow_definition,
            api_key=api_key,
            parameters=parameters,
            base_url=base_url,
            stream=True
        )
        return
    
    # Enhanced modes with event processing
    start_time = datetime.now()
    event_count = 0
    step_states = {}
    workflow_ended = False
    last_heartbeat = None
    validation_warnings = []
    
    # Pre-execution validation and setup
    if mode == ExecutionMode.LOGGING:
        workflow_name = workflow_definition.get('name', 'unknown')
        print(f"🚀 Executing Workflow: {workflow_name}")
        print("=" * 60)
        
        # Validate bounded services format and show warnings
        try:
            steps = workflow_definition.get('steps', [])
            print(f"📋 Steps: {len(steps)}")
            
            # Show services and validate them
            services = set()
            service_endpoints = {}
            for step_idx, step in enumerate(steps):
                executor = step.get('executor', {})
                if executor.get('type') == 'tool':
                    tool_def = executor.get('config', {}).get('tool_def', {})
                    step_services = tool_def.get('with_services', [])
                    
                    for service_idx, service in enumerate(step_services):
                        service_name = service.get('name', f'service-{service_idx}')
                        service_image = service.get('image', 'unknown')
                        
                        # Check for common validation issues and warn
                        if ':' not in service_image and not service_image.endswith(':latest'):
                            validation_warnings.append(f"Service '{service_name}' image '{service_image}' should include a tag")
                        
                        # Track service endpoints
                        ports = service.get('exposed_ports', [])
                        if ports:
                            endpoint = f"{service_name}-svc:{ports[0]}"
                            service_endpoints[service_name] = endpoint
                            services.add(f"{service_name} ({service_image}) -> {endpoint}")
                        else:
                            service_endpoints[service_name] = f"{service_name}-svc"
                            services.add(f"{service_name} ({service_image}) -> {service_name}-svc")
            
            if services:
                print(f"🔗 Bounded Services: {len(services)}")
                for service in sorted(services):
                    print(f"   • {service}")
                
                if log_level in [LogLevel.VERBOSE, LogLevel.DEBUG]:
                    print(f"\n🌐 Service Endpoints:")
                    for service_name, endpoint in service_endpoints.items():
                        print(f"   • {service_name} -> {endpoint}")
            
            # Show validation warnings
            if validation_warnings:
                print(f"\n⚠️  Validation Warnings:")
                for warning in validation_warnings:
                    print(f"   • {warning}")
            
            print()
            
        except Exception as e:
            if log_level in [LogLevel.VERBOSE, LogLevel.DEBUG]:
                print(f"⚠️  Could not analyze services: {str(e)}")
                print()
    
    try:
        # Execute workflow with comprehensive error handling
        for event in _execute_workflow_raw(
            workflow_definition=workflow_definition,
            api_key=api_key,
            parameters=parameters,
            base_url=base_url,
            stream=True
        ):
            event_count += 1
            current_time = datetime.now()
            elapsed = (current_time - start_time).total_seconds()
            
            # Parse event
            event_data = None
            try:
                event_data = json.loads(event)
            except json.JSONDecodeError:
                if mode == ExecutionMode.LOGGING and log_level == LogLevel.DEBUG:
                    if 'retry' not in event:
                        print(f"[{elapsed:6.1f}s] 📡 Raw: {event}")
                continue
            
            # Call user callback if provided
            if on_event and event_data:
                on_event(event_data)
            
            # Handle different event types
            event_type = event_data.get('type', 'unknown')
            
            if event_type == 'heartbeat':
                last_heartbeat = current_time
                if mode == ExecutionMode.LOGGING and log_level in [LogLevel.VERBOSE, LogLevel.DEBUG]:
                    print(f"[{elapsed:6.1f}s] 💓 Heartbeat")
                
            elif event_type == 'step_running':
                step_data = event_data.get('step', {})
                step_name = step_data.get('name', 'unknown')
                step_states[step_name] = 'running'
                
                if mode == ExecutionMode.LOGGING:
                    print(f"[{elapsed:6.1f}s] 🚀 Starting: {step_name}")
                
            elif event_type == 'step_finished':
                step_data = event_data.get('step', {})
                step_name = step_data.get('name', 'unknown')
                step_states[step_name] = 'finished'
                
                if mode == ExecutionMode.LOGGING:
                    print(f"[{elapsed:6.1f}s] ✅ Completed: {step_name}")
                
            elif 'output' in event_data:
                output = event_data['output']
                if mode == ExecutionMode.LOGGING:
                    # Filter and format output based on log level
                    if log_level == LogLevel.MINIMAL:
                        # Only show key results
                        if any(keyword in output.lower() for keyword in ['✅', '❌', 'error', 'failed', 'success']):
                            lines = output.split('\n')[:2]
                            preview = ' '.join(lines).strip()
                            if len(preview) > 80:
                                preview = preview[:80] + "..."
                            print(f"[{elapsed:6.1f}s] 📤 {preview}")
                    elif log_level == LogLevel.NORMAL:
                        # Show relevant output
                        lines = output.split('\n')
                        for line in lines[:3]:
                            line = line.strip()
                            if line and any(keyword in line for keyword in ['🚀', '✅', '❌', '🔗', 'service', 'endpoint']):
                                print(f"[{elapsed:6.1f}s] 📤 {line}")
                    elif log_level in [LogLevel.VERBOSE, LogLevel.DEBUG]:
                        # Show more detailed output
                        lines = output.split('\n')[:5]
                        for line in lines:
                            line = line.strip()
                            if line:
                                print(f"[{elapsed:6.1f}s] 📤 {line}")
                
            elif 'error' in event_data:
                error = event_data['error']
                if mode == ExecutionMode.LOGGING:
                    print(f"[{elapsed:6.1f}s] ❌ ERROR: {error}")
                
            elif event_data.get('end') or event_data.get('finishReason'):
                workflow_ended = True
                if mode == ExecutionMode.LOGGING:
                    print(f"[{elapsed:6.1f}s] 🏁 Workflow ended")
                
            # Yield based on mode
            if mode == ExecutionMode.EVENTS:
                # Add timing information to event
                event_data['_elapsed_seconds'] = elapsed
                event_data['_event_count'] = event_count
                yield event_data
            elif mode == ExecutionMode.LOGGING:
                # For logging mode, we mainly print but can still yield raw data
                yield event
            
            # Check if workflow ended
            if workflow_ended:
                break
    
    except Exception as e:
        # Enhanced error handling with helpful feedback
        if mode == ExecutionMode.LOGGING:
            print(f"\n❌ WORKFLOW EXECUTION ERROR")
            print("=" * 60)
            
            # Check for common error types and provide specific guidance
            error_str = str(e).lower()
            
            if "validation" in error_str and "servicespec" in error_str:
                print("🔍 Service Validation Error Detected!")
                print("   This error occurred while validating bounded services.")
                print("   Common issues:")
                print("   • Service name contains invalid characters")
                print("   • Service image missing tag (e.g., use 'redis:7-alpine' not 'redis')")
                print("   • Invalid port numbers (must be 1-65535)")
                print("   • Duplicate service names")
                print()
                print(f"📋 Error Details: {str(e)}")
                
            elif "validation" in error_str and "volume" in error_str:
                print("🔍 Volume Validation Error Detected!")
                print("   This error occurred while validating volume mounts.")
                print("   Common issues:")
                print("   • Paths must be absolute (start with '/')")
                print("   • Empty path strings")
                print()
                print(f"📋 Error Details: {str(e)}")
                
            elif "validation" in error_str and "tool" in error_str:
                print("🔍 Tool Validation Error Detected!")
                print("   This error occurred while validating tool definitions.")
                print("   Common issues:")
                print("   • Tool name or description is empty")
                print("   • Invalid tool type (must be docker, shell, python, etc.)")
                print("   • Docker tools missing image specification")
                print("   • Invalid arguments structure")
                print()
                print(f"📋 Error Details: {str(e)}")
                
            elif "unauthorized" in error_str or "authentication" in error_str:
                print("🔑 Authentication Error!")
                print("   Please check your API key:")
                print("   • Ensure KUBIYA_API_KEY is set correctly")
                print("   • Verify the API key hasn't expired")
                print("   • Check your organization permissions")
                print()
                print(f"📋 Error Details: {str(e)}")
                
            elif "timeout" in error_str:
                print("⏰ Timeout Error!")
                print("   The workflow execution timed out.")
                print("   This can happen when:")
                print("   • Services take too long to start up")
                print("   • Network connectivity issues")
                print("   • Heavy resource usage")
                print()
                if validation_warnings:
                    print("💡 Consider these validation warnings that might be related:")
                    for warning in validation_warnings:
                        print(f"   • {warning}")
                print()
                print(f"📋 Error Details: {str(e)}")
                
            else:
                print("🔍 General Execution Error")
                print(f"📋 Error Details: {str(e)}")
                
                if validation_warnings:
                    print("\n💡 There were validation warnings that might be related:")
                    for warning in validation_warnings:
                        print(f"   • {warning}")
            
            print("\n🛠️  Troubleshooting Tips:")
            print("   1. Check your workflow definition syntax")
            print("   2. Validate service configurations")
            print("   3. Ensure API key is valid and has permissions")
            print("   4. Try with a simpler workflow first")
            print("   5. Check network connectivity")
            
        # Re-raise the exception for the caller to handle
        raise
    
    finally:
        # Summary for logging mode
        if mode == ExecutionMode.LOGGING:
            total_time = (datetime.now() - start_time).total_seconds()
            print()
            print("=" * 60)
            print("🎉 Workflow Execution Complete")
            print("=" * 60)
            print(f"📊 Summary:")
            print(f"   • Total time: {total_time:.1f} seconds")
            print(f"   • Total events: {event_count}")
            print(f"   • Steps processed: {len(step_states)}")
            
            if last_heartbeat:
                heartbeat_age = (datetime.now() - last_heartbeat).total_seconds()
                print(f"   • Last heartbeat: {heartbeat_age:.1f} seconds ago")
            
            # Show step states
            if step_states and log_level in [LogLevel.NORMAL, LogLevel.VERBOSE, LogLevel.DEBUG]:
                print(f"\n📋 Step Results:")
                for step_name, state in step_states.items():
                    status_icon = "✅" if state == "finished" else "🔄" if state == "running" else "❓"
                    print(f"   {status_icon} {step_name}: {state}")


# Convenience functions for different modes
def execute_workflow_logged(
    workflow_definition: Union[Dict[str, Any], str],
    api_key: str,
    log_level: LogLevel = LogLevel.NORMAL,
    **kwargs
) -> Generator[str, None, None]:
    """Execute workflow with logging output."""
    return execute_workflow_with_logging(
        workflow_definition=workflow_definition,
        api_key=api_key,
        mode=ExecutionMode.LOGGING,
        log_level=log_level,
        **kwargs
    )


def execute_workflow_events(
    workflow_definition: Union[Dict[str, Any], str],
    api_key: str,
    on_event: Optional[Callable[[Dict[str, Any]], None]] = None,
    **kwargs
) -> Generator[Dict[str, Any], None, None]:
    """Execute workflow and yield structured event objects."""
    return execute_workflow_with_logging(
        workflow_definition=workflow_definition,
        api_key=api_key,
        mode=ExecutionMode.EVENTS,
        on_event=on_event,
        **kwargs
    )


def execute_workflow_raw(
    workflow_definition: Union[Dict[str, Any], str],
    api_key: str,
    **kwargs
) -> Generator[str, None, None]:
    """Execute workflow and yield raw SSE events."""
    return execute_workflow_with_logging(
        workflow_definition=workflow_definition,
        api_key=api_key,
        mode=ExecutionMode.RAW,
        **kwargs
    )


async def execute_workflow_with_validation(
    workflow_def: Dict[str, Any],
    parameters: Dict[str, Any],
    mode: ExecutionMode = ExecutionMode.EVENTS,
    log_level: LogLevel = LogLevel.NORMAL,
    api_token: Optional[str] = None,
    base_url: str = "https://api.kubiya.ai"
) -> AsyncGenerator[Dict[str, Any], None]:
    """
    Execute a workflow with validation and streaming feedback.
    
    This function provides enhanced execution capabilities specifically
    designed for MCP server integration, offering real-time feedback
    and comprehensive error handling.
    
    Args:
        workflow_def: Workflow definition dictionary
        parameters: Workflow parameters
        mode: Execution mode for output formatting
        log_level: Logging verbosity level
        api_token: Kubiya API token
        base_url: Kubiya API base URL
        
    Yields:
        Execution events with validation and progress information
    """
    execution_id = str(uuid.uuid4())
    
    try:
        # Emit start event
        yield {
            'type': 'execution_started',
            'execution_id': execution_id,
            'timestamp': datetime.now().isoformat(),
            'workflow': workflow_def.get('name', 'unnamed'),
            'parameters': parameters
        }
        
        # Try to validate workflow if validation is available
        try:
            from .validation import validate_workflow
            validation_result = validate_workflow(workflow_def)
        except ImportError:
            # Basic validation if the validation module isn't available
            validation_result = type('ValidationResult', (), {
                'valid': bool(workflow_def.get('steps')),
                'errors': [] if workflow_def.get('steps') else ['No steps defined'],
                'warnings': []
            })()
        
        if not validation_result.valid:
            yield {
                'type': 'validation_failed',
                'execution_id': execution_id,
                'timestamp': datetime.now().isoformat(),
                'errors': validation_result.errors,
                'warnings': validation_result.warnings
            }
            
            yield {
                'type': 'workflow_failed',
                'execution_id': execution_id,
                'timestamp': datetime.now().isoformat(),
                'error': f"Workflow validation failed: {', '.join(validation_result.errors)}"
            }
            return
        
        # Emit validation success
        yield {
            'type': 'validation_passed',
            'execution_id': execution_id,
            'timestamp': datetime.now().isoformat(),
            'warnings': validation_result.warnings
        }
        
        # Progress tracking
        total_steps = len(workflow_def.get('steps', []))
        current_step = 0
        
        # Execute workflow
        if not api_token:
            yield {
                'type': 'execution_failed',
                'execution_id': execution_id,
                'timestamp': datetime.now().isoformat(),
                'error': 'API token required for workflow execution',
                'details': 'Real workflow execution requires a valid Kubiya API token',
                'help': {
                    'message': 'To enable real execution, provide a valid API token',
                    'environment_variable': 'KUBIYA_API_TOKEN',
                    'example': 'export KUBIYA_API_TOKEN=your_token_here'
                }
            }
            
            yield {
                'type': 'workflow_failed',
                'execution_id': execution_id,
                'timestamp': datetime.now().isoformat(),
                'error': 'Cannot execute workflow without API token'
            }
            return
        
        # Real execution via Kubiya API
        try:
            from .client import StreamingKubiyaClient
        except ImportError:
            yield {
                'type': 'execution_failed',
                'execution_id': execution_id,
                'timestamp': datetime.now().isoformat(),
                'error': 'Kubiya client not available',
                'details': 'StreamingKubiyaClient could not be imported',
                'help': {
                    'message': 'Ensure the Kubiya Workflow SDK is properly installed',
                    'install_command': 'pip install kubiya-workflow-sdk'
                }
            }
            
            yield {
                'type': 'workflow_failed',
                'execution_id': execution_id,
                'timestamp': datetime.now().isoformat(),
                'error': 'Client dependency not available'
            }
            return
        
        client = StreamingKubiyaClient(api_token=api_token, base_url=base_url)
        
        try:
            async for event in client.execute_workflow_stream(
                workflow=workflow_def,
                params=parameters
            ):
                # Transform and forward events
                if mode == ExecutionMode.RAW:
                    yield event
                elif mode == ExecutionMode.EVENTS:
                    # Parse and structure events
                    structured_event = _structure_event(event, execution_id)
                    if structured_event:
                        yield structured_event
                        
                        # Update progress
                        if structured_event.get('type') == 'step_completed':
                            current_step += 1
                            yield {
                                'type': 'step_progress',
                                'execution_id': execution_id,
                                'timestamp': datetime.now().isoformat(),
                                'progress': {
                                    'current': current_step,
                                    'total': total_steps,
                                    'percentage': (current_step / total_steps * 100) if total_steps > 0 else 100
                                }
                            }
                elif mode == ExecutionMode.LOGGING:
                    # Enhanced logging format
                    log_event = _format_log_event(event, log_level)
                    if log_event:
                        yield log_event
            
            # Emit completion
            yield {
                'type': 'workflow_completed',
                'execution_id': execution_id,
                'timestamp': datetime.now().isoformat(),
                'outputs': {},  # Would extract from final event
                'duration_seconds': 0  # Would calculate from start time
            }
            
        except Exception as e:
            yield {
                'type': 'execution_error',
                'execution_id': execution_id,
                'timestamp': datetime.now().isoformat(),
                'error': str(e),
                'phase': 'execution',
                'help': {
                    'message': 'Workflow execution failed',
                    'common_causes': [
                        'Invalid API token',
                        'Network connectivity issues',
                        'Workflow validation errors',
                        'Service startup timeouts'
                    ]
                }
            }
            
            yield {
                'type': 'workflow_failed',
                'execution_id': execution_id,
                'timestamp': datetime.now().isoformat(),
                'error': str(e)
            }
            
    except Exception as e:
        yield {
            'type': 'unexpected_error',
            'execution_id': execution_id,
            'timestamp': datetime.now().isoformat(),
            'error': str(e),
            'traceback': traceback.format_exc()
        }
        
        yield {
            'type': 'workflow_failed',
            'execution_id': execution_id,
            'timestamp': datetime.now().isoformat(),
            'error': f'Unexpected error: {str(e)}'
        }


def _structure_event(raw_event: Dict[str, Any], execution_id: str) -> Optional[Dict[str, Any]]:
    """Structure raw events into standardized format."""
    # This would parse raw Kubiya API events and structure them
    # For now, return a basic structured event
    return {
        'type': 'raw_event',
        'execution_id': execution_id,
        'timestamp': datetime.now().isoformat(),
        'data': raw_event
    }


def _format_log_event(raw_event: Dict[str, Any], log_level: LogLevel) -> Optional[Dict[str, Any]]:
    """Format events for logging output."""
    # This would format events based on log level
    # For now, return a basic log event
    if log_level in [LogLevel.VERBOSE, LogLevel.DEBUG]:
        return {
            'type': 'log_event',
            'timestamp': datetime.now().isoformat(),
            'level': 'info',
            'message': f"Event: {raw_event.get('type', 'unknown')}",
            'data': raw_event if log_level == LogLevel.DEBUG else None
        }
    return None 